from tqdm.auto import tqdm
from pathlib import Path
import gzip
import dill as pickle
import functools

from stochastic_matching.simulator.metrics import get_metrics


def do_nothing(x):
    """

    Parameters
    ----------
    x: object
        Something

    Returns
    -------
    object
        Same object
    """
    return x


class Iterator:
    """
    Provides an easy way to make a parameter vary.

    Parameters
    ----------
    parameter: :class:`str`
        Name of the argument to vary.
    values: iterable
        Values that the argument can take.
    name: :class:`str`, optional
        *Display* name of the parameter
    process: callable, optional
        If you want to transform the value used, use this.


    Returns
    -------
    kwarg: :class:`dict`
        Keyword argument to use.
    log: :class:`dict`
        What you want to remind. By default, identical to kwarg.


    Examples
    --------

    Imagine one wants a parameter x2 to vary amongst squares of integers.

    >>> iterator = Iterator('x2', [i**2 for i in range(4)])
    >>> for kwarg, log in iterator:
    ...     print(kwarg, log)
    {'x2': 0} {'x2': 0}
    {'x2': 1} {'x2': 1}
    {'x2': 4} {'x2': 4}
    {'x2': 9} {'x2': 9}

    You can do the same thing by iterating over integers and apply a square method.

    >>> iterator = Iterator('x2', range(4), 'x', lambda x: x**2)
    >>> for kwarg, log in iterator:
    ...     print(kwarg, log)
    {'x2': 0} {'x': 0}
    {'x2': 1} {'x': 1}
    {'x2': 4} {'x': 2}
    {'x2': 9} {'x': 3}
    """

    def __init__(self, parameter, values, name=None, process=None):
        self.parameter = parameter
        self.values = values
        if name is None:
            name = parameter
        self.name = name
        if process is None:
            process = do_nothing
        self.process = process

    def __iter__(self):
        for v in self.values:
            yield {self.parameter: self.process(v)}, {self.name: v}

    def __len__(self):
        return len(self.values)


class SingleXP:
    """
    Produce a single experiment, with possibly one variable parameter.

    Parameters
    ----------
    name: :class:`str`
        Name of the experiment.
    iterator: :class:`~stochastic_matching.simulator.parallel.Iterator`, optional
        Argument that varies.
    params: :class:`dict`, optional
        Various (fixed) arguments.
    """

    def __init__(self, name, iterator=None, **params):
        self.name = name
        self.iterator = iterator
        self.params = params

    def __len__(self):
        if self.iterator is None:
            return 1
        else:
            return len(self.iterator)

    def __iter__(self):
        if self.iterator is None:
            yield self.name, None, self.params
        else:
            for (
                kwarg,
                log,
            ) in self.iterator:
                yield self.name, log, {**self.params, **kwarg}


class XP:
    """
    Produce an experiment, with possibly one variable parameter.
    Addition can be used to concatenate experiments.

    Parameters
    ----------
    name: :class:`str`
        Name of the experiment.
    iterator: :class:`~stochastic_matching.simulator.parallel.Iterator`, optional
        Argument that varies.
    params: :class:`dict`, optional
        Various (fixed) arguments.
    """

    def __init__(self, name, iterator=None, **params):
        self.xp_list = [SingleXP(name, iterator, **params)]

    def __iter__(self):
        for xp in self.xp_list:
            for x in xp:
                yield x

    def __add__(self, other):
        if other == 0:
            return self
        res = XP(name=None)
        res.xp_list = self.xp_list + other.xp_list
        return res

    def __radd__(self, other):
        return self + other

    def __len__(self):
        return sum(len(xp) for xp in self.xp_list)


class Runner:
    """
    Parameters
    ----------
    metrics: :class:`list`
        Metrics to extract.
    """

    def __init__(self, metrics):
        self.metrics = metrics

    def __call__(self, tup):
        name, kv, params = tup
        params = {**params}
        model = params.pop("model")
        model.run(**params)
        out = get_metrics(model.simulator, self.metrics)
        # logs = model.simulator.logs
        # out = {k: getattr(logs, k) for k in XP_LOGS}
        # if self.callback:
        #     out.update(self.callback(model))
        return name, kv, out


def aggregate(results):
    """
    Parameters
    ----------
    results: :class:`list`
        Computed results. Each entry is a dictionary associated to a given run.
    Returns
    -------
    :class:`dict`
        All results are gathered by experiment name, then by varying input / metric if applicable.
    """
    res = dict()
    for name, kv, r in results:
        if kv is None:
            res[name] = r
            continue
        rr = {**kv, **r}
        if name not in res:
            res[name] = {k: [] for k in rr}
        for k, v in rr.items():
            res[name][k].append(v)
    return res


def cached(func):
    """
    Enables functions that take a lot of time to compute to store their output on a file.

    If the original function is called with an additional parameter `cache_name`, it will look for a file named
    `cache_name.pkl.gz`. If it exists, it returns its content, otherwise it computes the results and
    stores it is returned.

    Additional parameter `cache_path` indicates where to look the file (default to current directory).

    Additional parameter `cache_overwrite` forces to always compute the function even if the file exists.
    Default to False.

    Parameters
    ----------
    func: callable
        Function to make cachable.

    Returns
    -------
    callable
        The function with three additional arguments: `cache_name`, `cache_path`, `cache_overwrite`.

    Examples
    --------

    >>> from tempfile import TemporaryDirectory as TmpDir
    >>> @cached
    ... def f(x):
    ...     return x**3
    >>> with TmpDir() as tmp:
    ...     content_before = [f.name for f in Path(tmp).glob('*')]
    ...     res = f(3, cache_name='cube', cache_path=Path(tmp))
    ...     content_after = [f.name for f in Path(tmp).glob('*')]
    ...     res2 = f(3, cache_name='cube', cache_path=Path(tmp)) # This does not run f!
    >>> res
    27
    >>> res2
    27
    >>> content_before
    []
    >>> content_after
    ['cube.pkl.gz']
    """

    @functools.wraps(func)
    def wrapper_decorator(
        *args, cache_name=None, cache_path=".", cache_overwrite=False, **kwargs
    ):
        if cache_name is not None:
            cache = Path(cache_path) / Path(f"{cache_name}.pkl.gz")
            if cache.exists() and not cache_overwrite:
                with gzip.open(cache, "rb") as f:
                    return pickle.load(f)
        else:
            cache = None
        res = func(*args, **kwargs)
        if cache is not None:
            with gzip.open(cache, "wb") as f:
                pickle.dump(res, f)
        return res

    return wrapper_decorator


@cached
def evaluate(xps, metrics=None, pool=None):
    """
    Evaluate some experiments. Results can be cached.

    Parameters
    ----------
    xps: :class:`~stochastic_matching.simulator.parallel.XP`
        Experiment(s) to run.
    metrics: :class:`list`, optional.
        Metrics to get.
    pool: :class:`~multiprocess.pool.Pool`, optional.
        Existing pool of workers.

    Returns
    -------
    :class:`dict`
        Result of the experiment(s).

    Examples
    --------

    >>> import stochastic_matching as sm
    >>> import numpy as np
    >>> diamond = sm.CycleChain()
    >>> base = {'model': diamond, 'n_steps': 1000, 'seed': 42, 'rewards': [1, 2.9, 1, -1, 1]}
    >>> xp = XP('Diamond', simulator='longest', **base)
    >>> res = evaluate(xp, ['income', 'steps_done', 'regret', 'delay'])
    >>> for name, r in res.items():
    ...     print(name)
    ...     for k, v in r.items():
    ...         print(f"{k}: {np.round(v, 4)}")
    Diamond
    income: [216 304 293 187]
    steps_done: 1000
    regret: 0.088
    delay: 0.1634
    >>> def q0(simu):
    ...     return simu.avg_queues[0]
    >>> res = evaluate(xp, ['flow', 'avg_queues', q0])
    >>> res['Diamond']['flow']
    array([1.19, 0.97, 0.97, 0.88, 0.99])
    >>> res['Diamond']['avg_queues']
    array([0.531, 0.214, 0.273, 0.616])
    >>> res['Diamond']['q0']
    np.float64(0.531)
    >>> xp1 = XP('e-filtering', simulator='e_filtering', **base,
    ...          iterator=Iterator('epsilon', [.01, .1, 1], name='e'))
    >>> xp2 = XP(name='k-filtering', simulator='longest', forbidden_edges=True,
    ...          iterator=Iterator('k', [0, 10, 100]), **base)
    >>> xp3 = XP(name='egpd', simulator='virtual_queue',
    ...          iterator=Iterator('beta', [.01, .1, 1]), **base)
    >>> xp = sum([xp1, xp2, xp3])
    >>> len(xp)
    9
    >>> import multiprocess as mp
    >>> with mp.Pool(processes=2) as p:
    ...     res = evaluate(xp, ['regret', 'delay'], pool=p)
    >>> for name, r in res.items():
    ...     print(name)
    ...     for k, v in r.items():
    ...         print(f"{k}: {np.array(v)}")
    e-filtering
    e: [0.01 0.1  1.  ]
    regret: [0.002 0.017 0.103]
    delay: [1.0538 0.695  0.1952]
    k-filtering
    k: [  0  10 100]
    regret: [ 8.8000000e-02  2.0000000e-03 -8.8817842e-16]
    delay: [0.1634 0.7342 1.3542]
    egpd
    beta: [0.01 0.1  1.  ]
    regret: [0.003 0.043 0.076]
    delay: [9.2024 1.013  0.1888]
    """
    if metrics is None:
        metrics = ["regret", "delay"]
    runner = Runner(metrics=metrics)
    if pool is None:
        res = [runner(p) for p in tqdm(xps)]
    else:
        res = tqdm(pool.imap(runner, xps), total=len(xps))
    return aggregate(res)



import numpy as np
from numba import njit

from stochastic_matching.simulator.simulator import Simulator
from stochastic_matching.simulator.multiqueue import MultiQueue


@njit
def lin_fading(t):
    """

    Parameters
    ----------
    t: :class:`int`
        Step index.

    Returns
    -------
    :class:`int`
        Step index + 1
    """
    return t + 1


@njit
def cr_core(
    # Generic arguments
    logs,
    arrivals,
    graph,
    n_steps,
    queue_size,
    # Constant-regret specific arguments
    scores,
    ready_edges,
    edge_queue,
    fading
):
    """
    Jitted function for constant-regret policy.

    Parameters
    ----------
    logs: :class:`~stochastic_matching.simulator.logs.Logs`
        Monitored variables.
    arrivals: :class:`~stochastic_matching.simulator.arrivals.Arrivals`
        Item arrival process.
    graph: :class:`~stochastic_matching.simulator.graph.JitHyperGraph`
        Model graph.
    n_steps: :class:`int`
        Number of arrivals to process.
    queue_size: :class:`~numpy.ndarray`
        Number of waiting items of each class.
    scores: :class:`~numpy.ndarray`
        Normalized edge rewards (<0 on taboo edges). Enables EGPD-like selection mechanism.
    ready_edges: :class:`~numpy.ndarray`
        Tells if given edge is actionable.
    edge_queue: :class:`~stochastic_matching.simulator.multiqueue.MultiQueue`
        Edges waiting to be matched.
    fading: callable
        Jitted function that drives reward relative weight.

    Returns
    -------
    None
    """
    n, max_queue = logs.queue_log.shape
    m = len(scores)

    infinity = edge_queue.infinity

    edge_size = np.zeros(m, dtype=np.int32)

    for age in range(n_steps):
        # Draw an arrival
        node = arrivals.draw()
        queue_size[node] += 1
        if queue_size[node] == max_queue:
            return None

        # update scores
        edge_size[graph.edges(node)] += 1

        # update readyness
        if queue_size[node] == 1:
            for e in graph.edges(node):
                for v in graph.nodes(e):
                    if v != node and queue_size[v] == 0:
                        ready_edges[e] = False
                        break
                else:
                    ready_edges[e] = True

        # Select best edge
        best_edge = -1
        best_score = 0

        for e in range(m):
            score = scores[e] + edge_size[e] / fading(age)
            if score > best_score:
                best_edge = e
                best_score = score

        if best_edge > -1:
            # add edge to virtual queue
            edge_queue.add(best_edge, age)
            # Virtual pop of items:
            # for each node of the edge, lower scores of all adjacent edges by one
            for v in graph.nodes(best_edge):
                edge_size[graph.edges(v)] -= 1

        # Can a virtual edge be popped?
        best_edge = -1
        best_age = infinity
        for e in range(m):
            if ready_edges[e]:
                edge_age = edge_queue.oldest(e)
                if edge_age < best_age:
                    best_edge = e
                    best_age = edge_age
        if best_edge > -1:
            edge_queue.pop(best_edge)
            for v in graph.nodes(best_edge):
                queue_size[v] -= 1
                if queue_size[v] == 0:
                    ready_edges[graph.edges(v)] = False

        logs.update(queue_size=queue_size, node=node, edge=best_edge)
    return None


class ConstantRegret(Simulator):
    """
    Non-Greedy Matching simulator that implements https://sophieyu.me/constant_regret_matching.pdf
    Always pick-up the best positive edge according to a scoring function, even if that edge cannot be used (yet).

    Parameters
    ----------
    model: :class:`~stochastic_matching.model.Model`
        Model to simulate.
    rewards: :class:`~numpy.ndarray`
        Edge rewards.
    fading: callable
        Jitted function that drives reward relative weight.
    max_edge_queue: :class:`int`, optional
        In some extreme situation, the default allocated space for the edge virtual queue may be too small.
        If that happens someday, use this parameter to increase the VQ allocated memory.
    **kwargs
        Keyword parameters of :class:`~stochastic_matching.simulator.simulator.Simulator`.


    Examples
    --------

    Let start with a working triangle.

    >>> import stochastic_matching as sm
    >>> sim = ConstantRegret(sm.Cycle(rates=[3, 4, 5]), n_steps=1000, seed=42, max_queue=10)
    >>> sim.run()
    >>> sim.plogs # doctest: +NORMALIZE_WHITESPACE
    Arrivals: [287 338 375]
    Traffic: [125 162 213]
    Queues: [[837 105  41  13   3   1   0   0   0   0]
     [784 131  53  22   8   2   0   0   0   0]
     [629 187  92  51  24   9   5   3   0   0]]
    Steps done: 1000

    Unstable diamond (simulation ends before completion due to drift).

    >>> sim = ConstantRegret(sm.CycleChain(rates='uniform'), n_steps=1000, seed=42, max_queue=10)
    >>> sim.run()
    >>> sim.plogs # doctest: +NORMALIZE_WHITESPACE
    Arrivals: [85 82 85 86]
    Traffic: [34 42  7 41 36]
    Queues: [[141  55  34  25  23  16  24  12   7   1]
     [318  16   3   1   0   0   0   0   0   0]
     [316  17   4   1   0   0   0   0   0   0]
     [106  67  64  37  22  24   7   3   6   2]]
    Steps done: 338

    Stable diamond without reward optimization:

    >>> sim = ConstantRegret(sm.CycleChain(rates=[1, 2, 2, 1]), n_steps=1000, seed=42, max_queue=10)
    >>> sim.run()
    >>> sim.plogs # doctest: +NORMALIZE_WHITESPACE
    Arrivals: [179 342 320 159]
    Traffic: [ 95  84 161  84  75]
    Queues: [[823 120  38  19   0   0   0   0   0   0]
     [625 215  76  46  28   9   1   0   0   0]
     [686 212  70  24   7   1   0   0   0   0]
     [823 118  39  12   4   4   0   0   0   0]]
    Steps done: 1000

    Let's optimize (kill traffic on first and last edges).

    >>> sim = ConstantRegret(sm.CycleChain(rates=[1, 2, 2, 1]), rewards=[0, 1, 1, 1, 0],
    ...                    n_steps=1000, seed=42, max_queue=10)
    >>> sim.run()
    >>> sim.plogs # doctest: +NORMALIZE_WHITESPACE
    Arrivals: [20 44 51 23]
    Traffic: [ 3 17 25 16  0]
    Queues: [[136   2   0   0   0   0   0   0   0   0]
     [127   4   6   1   0   0   0   0   0   0]
     [ 24  22   9  13  14  20  16  13   4   3]
     [ 40  17  22  21  22  14   1   1   0   0]]
    Steps done: 138

    OK, it's mostly working, but we reached the maximal queue size quite fast. Let's reduce the pressure.

    >>> slow_fading = njit(lambda t: np.sqrt(t+1)/15)
    >>> sim = ConstantRegret(sm.CycleChain(rates=[1, 2, 2, 1]), rewards=[0, 1, 1, 1, 0],
    ...                    fading=slow_fading, n_steps=1000, seed=42, max_queue=10)
    >>> sim.run()
    >>> sim.plogs # doctest: +NORMALIZE_WHITESPACE
    Arrivals: [179 342 320 159]
    Traffic: [ 43 136 161 136  23]
    Queues: [[900  79  18   2   1   0   0   0   0   0]
     [840  98  37  17   6   2   0   0   0   0]
     [483 224 122  78  61  26   5   1   0   0]
     [551 255 100  50  25  10   9   0   0   0]]
    Steps done: 1000

    A stable candy. While candies are not good for greedy policies, virtual queues policies are
    designed to deal with it.

    >>> sim = ConstantRegret(sm.HyperPaddle(rates=[1, 1, 1.5, 1, 1.5, 1, 1]), n_steps=1000, seed=42, max_queue=25)
    >>> sim.run()
    >>> sim.plogs # doctest: +NORMALIZE_WHITESPACE
    Arrivals: [140 126 154 112 224 121 123]
    Traffic: [109  29  17  59  58  62 107]
    Queues: [[301  83  94  83  54  60  43  48  41  32  49  60  31   3   8   7   3   0
        0   0   0   0   0   0   0]
     [825 101  27  14  26   7   0   0   0   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]
     [899  64  20   8   9   0   0   0   0   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]
     [110  49  93 108 120 102  48  28  53  32   7  30  16  42  28  34  31  30
       26  11   2   0   0   0   0]
     [276 185 151  95  60  40  49  52  47  33  12   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]
     [755 142  67  36   0   0   0   0   0   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]
     [709 228  47  13   3   0   0   0   0   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]]
    Steps done: 1000

    Last but not least: Stolyar's example from https://arxiv.org/abs/1608.01646

    >>> ns = sm.NS19(rates=[1.2, 1.5, 2, .8])

    Without optimization, all we do is self-matches (no queue at all):

    >>> sim = ConstantRegret(ns, n_steps=1000, seed=42, max_queue=25)
    >>> sim.run()
    >>> sim.logs.traffic.astype(int)
    array([236, 279, 342, 143,   0,   0,   0])
    >>> sim.avg_queues
    array([0., 0., 0., 0.])

    Let's introduce some rewards:

    >>> rewards = [-1, -1, 1, 2, 5, 4, 7]
    >>> ns.optimize_rates(rewards)
    array([0. , 0. , 1.7, 0.5, 1.2, 0. , 0.3])
    >>> ns.normalize_rewards(rewards)
    array([-2., -5.,  0.,  0.,  0., -1.,  0.])

    With optimization, we get the desired results:

    >>> sim = ConstantRegret(ns, rewards=rewards, n_steps=3000, seed=42, max_queue=300)
    >>> sim.run()
    >>> sim.logs.traffic.astype(int)
    array([  0,   0, 961, 342, 691,   0, 105])
    >>> sim.avg_queues
    array([2.731     , 0.69633333, 0.22266667, 0.052     ])

    One can check that this is the same as using a regular virtual queue with forbidden edges:

    >>> from stochastic_matching.simulator.virtual_queue import VirtualQueue
    >>> sim = VirtualQueue(ns, rewards=rewards, forbidden_edges=True, n_steps=3000, seed=42, max_queue=300)
    >>> sim.run()
    >>> sim.logs.traffic.astype(int)
    array([  0,   0, 961, 342, 691,   0, 105])
    >>> sim.avg_queues
    array([2.731     , 0.69633333, 0.22266667, 0.052     ])
    """

    name = "constant_regret"

    def __init__(self, model, rewards=None, fading=None, max_edge_queue=None, **kwargs):

        if rewards is not None:
            rewards = np.array(rewards)
        else:
            rewards = model.incidence.sum(axis=0)
        self.rewards = model.normalize_rewards(rewards)

        if fading is None:
            fading = lin_fading
        self.fading = fading

        self.max_edge_queue = max_edge_queue

        super().__init__(model, **kwargs)

    def set_internal(self):
        super().set_internal()
        self.internal["scores"] = self.rewards
        self.internal["fading"] = self.fading
        self.internal["ready_edges"] = np.zeros(self.model.m, dtype=np.bool_)
        meq = self.max_edge_queue
        if meq is None:
            meq = 10 * self.max_queue
        self.internal["edge_queue"] = MultiQueue(self.model.m, max_queue=meq)

    def run(self):
        cr_core(logs=self.logs, **self.internal)



from numba import njit
import numpy as np

from stochastic_matching.simulator.extended import ExtendedSimulator
from stochastic_matching.simulator.multiqueue import MultiQueue


@njit
def vq_core(
    logs,
    arrivals,
    graph,
    n_steps,
    queue_size,  # Generic arguments
    scores,
    ready_edges,
    edge_queue,
    forbidden_edges,
    k,
):  # Longest specific arguments
    """
    Jitted function for virtual-queue policy.

    Parameters
    ----------
    logs: :class:`~stochastic_matching.simulator.logs.Logs`
        Monitored variables.
    arrivals: :class:`~stochastic_matching.simulator.arrivals.Arrivals`
        Item arrival process.
    graph: :class:`~stochastic_matching.simulator.graph.JitHyperGraph`
        Model graph.
    n_steps: :class:`int`
        Number of arrivals to process.
    queue_size: :class:`~numpy.ndarray`
        Number of waiting items of each class.
    scores: :class:`~numpy.ndarray`
        Default value for edges. Enables EGPD-like selection mechanism.
    ready_edges: :class:`~numpy.ndarray`
        Tells if given edge is actionable.
    edge_queue: :class:`~stochastic_matching.simulator.multiqueue.MultiQueue`
        Edges waiting to be matched.
    forbidden_edges: :class:`list`, optional
        Edges that are disabled.
    k: :class:`int`, optional
        Queue size above which forbidden edges become available again.

    Returns
    -------
    None
    """

    n, max_queue = logs.queue_log.shape
    m = len(scores)
    if k is None:
        k = max_queue + 1

    # Optimize forbidden edges and set greedy flag.
    if forbidden_edges is not None:
        forbid = {k: True for k in forbidden_edges}
        has_forbidden = True
    else:
        forbid = {k: True for k in range(0)}
        has_forbidden = False

    infinity = edge_queue.infinity

    for age in range(n_steps):
        # Draw an arrival
        node = arrivals.draw()
        queue_size[node] += 1
        if queue_size[node] == max_queue:
            return None

        # update scores
        scores[graph.edges(node)] += 1

        # update readyness
        if queue_size[node] == 1:
            for e in graph.edges(node):
                for v in graph.nodes(e):
                    if v != node and queue_size[v] == 0:
                        ready_edges[e] = False
                        break
                else:
                    ready_edges[e] = True

        if has_forbidden:
            restrain = True
            if k is not None:
                for v in range(n):
                    if queue_size[v] >= k:
                        restrain = False
                        break
        else:
            restrain = False

        # Select best edge
        best_edge = -1
        best_score = 0

        for e in range(m):
            if restrain and e in forbid:
                continue
            score = scores[e]
            if score > best_score:
                best_edge = e
                best_score = score

        if best_edge > -1:
            # add edge to virtual queue
            edge_queue.add(best_edge, age)
            # Virtual pop of items:
            # for each node of the edge, lower scores of all adjacent edges by one
            for v in graph.nodes(best_edge):
                scores[graph.edges(v)] -= 1

        # Can a virtual edge be popped?
        best_edge = -1
        best_age = infinity
        for e in range(m):
            if ready_edges[e]:
                edge_age = edge_queue.oldest(e)
                if edge_age < best_age:
                    best_edge = e
                    best_age = edge_age
        if best_edge > -1:
            edge_queue.pop(best_edge)
            for v in graph.nodes(best_edge):
                queue_size[v] -= 1
                if queue_size[v] == 0:
                    ready_edges[graph.edges(v)] = False

        logs.update(queue_size=queue_size, node=node, edge=best_edge)


class VirtualQueue(ExtendedSimulator):
    """
    Non-Greedy Matching simulator derived from :class:`~stochastic_matching.simulator.extended.ExtendedSimulator`.
    Always pick-up the best edge according to a scoring function, even if that edge cannot be used (yet).

    Parameters
    ----------
    model: :class:`~stochastic_matching.model.Model`
        Model to simulate.
    max_edge_queue: :class:`int`, optional
        In some extreme situation, the default allocated space for the edge virtual queue may be too small.
        If that happens someday, use this parameter to increase the VQ allocated memory.
    **kwargs
        Keyword parameters of :class:`~stochastic_matching.simulator.extended.ExtendedSimulator`.


    Examples
    --------

    Let start with a working triangle.

    >>> import stochastic_matching as sm
    >>> sim = VirtualQueue(sm.Cycle(rates=[3, 4, 5]), n_steps=1000, seed=42, max_queue=10)
    >>> sim.run()
    >>> sim.plogs # doctest: +NORMALIZE_WHITESPACE
    Arrivals: [287 338 375]
    Traffic: [125 162 213]
    Queues: [[837 105  41  13   3   1   0   0   0   0]
     [784 131  53  22   8   2   0   0   0   0]
     [629 187  92  51  24   9   5   3   0   0]]
    Steps done: 1000

    Unstable diamond (simulation ends before completion due to drift).

    >>> sim = VirtualQueue(sm.CycleChain(rates='uniform'), n_steps=1000, seed=42, max_queue=10)
    >>> sim.run()
    >>> sim.plogs # doctest: +NORMALIZE_WHITESPACE
    Arrivals: [85 82 85 86]
    Traffic: [34 42  7 41 36]
    Queues: [[141  55  34  25  23  16  24  12   7   1]
     [318  16   3   1   0   0   0   0   0   0]
     [316  17   4   1   0   0   0   0   0   0]
     [106  67  64  37  22  24   7   3   6   2]]
    Steps done: 338

    Stable diamond without reward optimization:

    >>> sim = VirtualQueue(sm.CycleChain(rates=[1, 2, 2, 1]), n_steps=1000, seed=42, max_queue=10)
    >>> sim.run()
    >>> sim.plogs # doctest: +NORMALIZE_WHITESPACE
    Arrivals: [179 342 320 159]
    Traffic: [ 95  84 161  84  75]
    Queues: [[823 120  38  19   0   0   0   0   0   0]
     [625 215  76  46  28   9   1   0   0   0]
     [686 212  70  24   7   1   0   0   0   0]
     [823 118  39  12   4   4   0   0   0   0]]
    Steps done: 1000

    Let's optimize (kill traffic on first and last edges).

    >>> sim = VirtualQueue(sm.CycleChain(rates=[1, 2, 2, 1]), rewards=[0, 1, 1, 1, 0], beta=.01,
    ...                    n_steps=1000, seed=42, max_queue=10)
    >>> sim.run()
    >>> sim.plogs # doctest: +NORMALIZE_WHITESPACE
    Arrivals: [22 51 55 27]
    Traffic: [ 0 22 24 27  0]
    Queues: [[136  16   3   0   0   0   0   0   0   0]
     [112  14   8   6   7   4   3   1   0   0]
     [ 73  21  13   7   9   8   4   5  12   3]
     [105  31  11   5   2   1   0   0   0   0]]
    Steps done: 155

    OK, it's working, but we reached the maximal queue size quite fast. Let's reduce the pressure.

    >>> sim = VirtualQueue(sm.CycleChain(rates=[1, 2, 2, 1]), rewards=[0, 1, 1, 1, 0],
    ...                    beta=.8, n_steps=1000, seed=42, max_queue=10)
    >>> sim.run()
    >>> sim.plogs # doctest: +NORMALIZE_WHITESPACE
    Arrivals: [179 342 320 159]
    Traffic: [ 32 146 161 146  13]
    Queues: [[691 222  68  18   1   0   0   0   0   0]
     [514 153 123 110  48  32  16   4   0   0]
     [662 136  91  62  38   4   2   2   2   1]
     [791 128  50  18   7   6   0   0   0   0]]
    Steps done: 1000

    Let's play a bit with alt rewards, starting with adversarial rewards:

    >>> rewards = [-1, 1, 1, 1, 2.9]
    >>> sim = VirtualQueue(sm.CycleChain(rates=[1, 2, 2, 1]), rewards=rewards,
    ...                    beta=.8, n_steps=1000, seed=42, max_queue=10)
    >>> sim.run()
    >>> sim.plogs # doctest: +NORMALIZE_WHITESPACE
    Arrivals: [179 342 320 159]
    Traffic: [ 70 109 161 109  50]
    Queues: [[664 213  84  28   8   3   0   0   0   0]
     [460 247 135  70  50  25  10   3   0   0]
     [824 122  38  10   5   1   0   0   0   0]
     [902  71  20   7   0   0   0   0   0   0]]
    Steps done: 1000

    With those rewards, we are far from the target. Let's be gentle:

    >>> sim = VirtualQueue(sm.CycleChain(rates=[1, 2, 2, 1]), rewards=rewards,
    ...                    alt_rewards="gentle", beta=.8, n_steps=1000, seed=42, max_queue=10)
    >>> sim.model.gentle_rewards(rewards)
    array([-2,  2,  2,  2, -2])
    >>> sim.run()
    >>> sim.plogs # doctest: +NORMALIZE_WHITESPACE
    Arrivals: [29 58 68 32]
    Traffic: [ 0 29 29 29  1]
    Queues: [[168  16   3   0   0   0   0   0   0   0]
     [143  15   9   5   7   4   3   1   0   0]
     [ 70  24  16   9  22  18  11   8   4   5]
     [ 95  32  23  10  19   7   1   0   0   0]]
    Steps done: 187

    OK, it works but pressure is restored. Let's normalize:

    >>> sim = VirtualQueue(sm.CycleChain(rates=[1, 2, 2, 1]), rewards=rewards,
    ...                    alt_rewards="normalize", beta=.8, n_steps=1000, seed=42, max_queue=10)
    >>> sim.model.normalize_rewards(rewards)
    array([ 0. ,  0. ,  0. ,  0. , -0.1])
    >>> sim.run()
    >>> sim.plogs # doctest: +NORMALIZE_WHITESPACE
    Arrivals: [179 342 320 159]
    Traffic: [ 95  84 161  84  75]
    Queues: [[823 120  38  19   0   0   0   0   0   0]
     [625 215  76  46  28   9   1   0   0   0]
     [686 212  70  24   7   1   0   0   0   0]
     [823 118  39  12   4   4   0   0   0   0]]
    Steps done: 1000

    Not enough pressure in this example!

    Let's now use a k-filtering approach for this diamond:

    >>> sim = VirtualQueue(sm.CycleChain(rates=[1, 2, 2, 1]), forbidden_edges=True,
    ...                    rewards=[0, 1, 1, 1, 0], n_steps=1000, seed=42, max_queue=10)
    >>> sim.run()
    >>> sim.plogs # doctest: +NORMALIZE_WHITESPACE
    Arrivals: [27 57 65 31]
    Traffic: [ 0 27 29 28  0]
    Queues: [[158  16   3   3   0   0   0   0   0   0]
     [136  19   8   4   5   4   3   1   0   0]
     [ 70  25  13  14  12  15  21   8   1   1]
     [ 96  24  13  12   9  16   8   2   0   0]]
    Steps done: 180

    Let us reduce the pressure.

    >>> sim = VirtualQueue(sm.CycleChain(rates=[1, 2, 2, 1]), forbidden_edges=True, rewards=[0, 1, 1, 1, 0],
    ...                    n_steps=1000, seed=42, max_queue=10, k=6)
    >>> sim.run()
    >>> sim.plogs # doctest: +NORMALIZE_WHITESPACE
    Arrivals: [179 342 320 159]
    Traffic: [ 31 145 161 145  14]
    Queues: [[600 171 122  81  20   6   0   0   0   0]
     [444 151 137 104  81  65  14   4   0   0]
     [698  83  52  59  57  38   7   3   2   1]
     [730 130  64  35  26   9   2   4   0   0]]
    Steps done: 1000

    A stable candy. While candies are not good for greedy policies, the virtual queue is
    designed to deal with it.

    >>> sim = VirtualQueue(sm.HyperPaddle(rates=[1, 1, 1.5, 1, 1.5, 1, 1]), n_steps=1000, seed=42, max_queue=25)
    >>> sim.run()
    >>> sim.plogs # doctest: +NORMALIZE_WHITESPACE
    Arrivals: [140 126 154 112 224 121 123]
    Traffic: [109  29  17  59  58  62 107]
    Queues: [[301  83  94  83  54  60  43  48  41  32  49  60  31   3   8   7   3   0
        0   0   0   0   0   0   0]
     [825 101  27  14  26   7   0   0   0   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]
     [899  64  20   8   9   0   0   0   0   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]
     [110  49  93 108 120 102  48  28  53  32   7  30  16  42  28  34  31  30
       26  11   2   0   0   0   0]
     [276 185 151  95  60  40  49  52  47  33  12   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]
     [755 142  67  36   0   0   0   0   0   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]
     [709 228  47  13   3   0   0   0   0   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]]
    Steps done: 1000

    Last but not least: Stolyar's example from https://arxiv.org/abs/1608.01646

    >>> stol = sm.Model(incidence=[[1, 0, 0, 0, 1, 0, 0],
    ...                       [0, 1, 0, 0, 1, 1, 1],
    ...                       [0, 0, 1, 0, 0, 1, 1],
    ...                       [0, 0, 0, 1, 0, 0, 1]], rates=[1.2, 1.5, 2, .8])

    Without optimization, all we do is self-matches (no queue at all):

    >>> sim = VirtualQueue(stol, n_steps=1000, seed=42, max_queue=25)
    >>> sim.run()
    >>> sim.logs.traffic.astype(int)
    array([236, 279, 342, 143,   0,   0,   0])
    >>> sim.avg_queues
    array([0., 0., 0., 0.])

    >>> rewards = [-1, -1, 1, 2, 5, 4, 7]

    With optimization, we get the desired results at the price of a huge queue:

    >>> sim = VirtualQueue(stol, rewards=rewards, beta=.01, n_steps=3000, seed=42, max_queue=300)
    >>> sim.run()
    >>> sim.logs.traffic.astype(int)
    array([  0,   0, 761, 242, 591,   0, 205])
    >>> sim.avg_queues
    array([86.17266667,  0.53233333, 93.03      ,  0.33533333])

    Same traffic could be achieved with much lower queues by enforcing forbidden edges:

    >>> sim = VirtualQueue(stol, rewards=rewards, forbidden_edges=True, n_steps=3000, seed=42, max_queue=300)
    >>> sim.run()
    >>> sim.logs.traffic.astype(int)
    array([  0,   0, 961, 342, 691,   0, 105])
    >>> sim.avg_queues
    array([2.731     , 0.69633333, 0.22266667, 0.052     ])
    """

    name = "virtual_queue"

    def __init__(self, model, max_edge_queue=None, **kwargs):
        self.max_edge_queue = max_edge_queue
        super().__init__(model, **kwargs)

    def set_internal(self):
        super().set_internal()
        self.internal["ready_edges"] = np.zeros(self.model.m, dtype=np.bool_)
        meq = self.max_edge_queue
        if meq is None:
            meq = 10 * self.max_queue
        self.internal["edge_queue"] = MultiQueue(self.model.m, max_queue=meq)

    def run(self):
        vq_core(logs=self.logs, **self.internal)



from inspect import isclass

import numpy as np
from numba import njit


def pseudo_inverse_scalar(x):
    """
    Parameters
    ----------
    x: :class:`float`

    Returns
    -------
    :class:`float`
        Inverse of x if it is not 0.

    Examples
    --------

    >>> pseudo_inverse_scalar(2.0)
    0.5
    >>> pseudo_inverse_scalar(0)
    0.0
    """
    return 0.0 if x == 0 else 1 / x


def clean_zeros(matrix, tol=1e-10):
    """
    Replace in-place all small values of a matrix by 0.

    Parameters
    ----------
    matrix: :class:`~numpy.ndarray`
        Matrix to clean.
    tol: :class:`float`, optional
        Threshold. All entries with absolute value lower than `tol` are put to zero.

    Returns
    -------
    None

    Examples
    --------

    >>> import numpy as np
    >>> mat = np.array([[1e-12, -.3], [.8, -1e-13]])
    >>> clean_zeros(mat)
    >>> mat # doctest: +NORMALIZE_WHITESPACE
    array([[ 0. , -0.3],
           [ 0.8,  0. ]])
    """
    matrix[abs(matrix[:]) < tol] = 0


class CharMaker:
    """
    Class that acts as an infinite list of letters. Used to provide letter-labels to nodes

    Examples
    --------

    >>> names = CharMaker()
    >>> names[0]
    'A'
    >>> names[7]
    'H'
    >>> names[26]
    'AA'
    >>> names[107458610947716]
    'STOCHASTIC'
    """

    def __init__(self):
        pass

    @staticmethod
    def to_char(i):
        return chr(ord("A") + (i % 26))

    def __getitem__(self, i):
        res = self.to_char(i)
        while i > 25:
            i = i // 26 - 1
            res = f"{self.to_char(i)}{res}"
        return res


def get_classes(root):
    """
    Parameters
    ----------
    root: :class:`class`
        Starting class (can be abstract).

    Returns
    -------
    :class:`dict`
        Dictionaries of all subclasses that have a name (as in class attribute `name`).

    Examples
    --------
    >>> import stochastic_matching as sm
    >>> get_classes(sm.Model) # doctest: +NORMALIZE_WHITESPACE
    {'Path': <class 'stochastic_matching.graphs.Path'>,
     'Star': <class 'stochastic_matching.graphs.Star'>,
     'Cycle': <class 'stochastic_matching.graphs.Cycle'>,
     'Complete': <class 'stochastic_matching.graphs.Complete'>,
     'Codomino': <class 'stochastic_matching.graphs.Codomino'>,
     'Erdös-Rényi': <class 'stochastic_matching.graphs.ErdosRenyi'>,
     'Nazari-Stolyar': <class 'stochastic_matching.graphs.NS19'>,
     'Pyramid': <class 'stochastic_matching.graphs.Pyramid'>,
     'Tadpole': <class 'stochastic_matching.graphs.Tadpole'>,
     'Lollipop': <class 'stochastic_matching.graphs.Lollipop'>,
     'Kayak Paddle': <class 'stochastic_matching.graphs.KayakPaddle'>,
     'Barbell': <class 'stochastic_matching.graphs.Barbell'>,
     'Cycle Chain': <class 'stochastic_matching.graphs.CycleChain'>,
     'Hyper Kayak Paddle': <class 'stochastic_matching.graphs.HyperPaddle'>,
     'Fan': <class 'stochastic_matching.graphs.Fan'>}
    """
    result = {c.name: c for c in root.__subclasses__() if c.name}
    for c in root.__subclasses__():
        result.update(get_classes(c))
    return result


def class_converter(subclass, motherclass):
    """
    Parameters
    ----------
    subclass: :class:`str` or :class:`class`
        Required subclass, or its name.
    motherclass: :class:`class`
        Ancestor of the subclass.

    Returns
    -------
    :class:`class`
        The subclass.

    Examples
    --------

    >>> import stochastic_matching as sm
    >>> from stochastic_matching.simulator.simulator import Simulator
    >>> class_converter('random_edge', Simulator)
    <class 'stochastic_matching.simulator.random_edge.RandomEdge'>
    >>> class_converter(sm.FCFM, Simulator)
    <class 'stochastic_matching.simulator.fcfm.FCFM'>
    >>> class_converter('anything', Simulator)  # doctest: +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
    ...
    ValueError: anything is not a known name for a subclass of Simulator.
    Known names: constant_regret, e_filtering, fcfm, longest, priority, random_edge, random_item, virtual_queue.

    >>> class_converter(2, Simulator)
    Traceback (most recent call last):
    ...
    TypeError: Subclass must be string or Simulator subclass (not instance).
    """
    if isinstance(subclass, str):
        class_dict = get_classes(motherclass)
        if subclass in class_dict:
            return class_dict[subclass]
        else:
            raise ValueError(
                f"{subclass} is not a known name for a subclass of {motherclass.__name__}. "
                f"Known names: {', '.join(sorted(class_dict.keys()))}."
            )
    elif isclass(subclass) and issubclass(subclass, motherclass):
        return subclass
    else:
        raise TypeError(
            f"Subclass must be string or {motherclass.__name__} subclass (not instance)."
        )


def neighbors(i, compressed_incidence):
    """
    Return neighbors of a node/edge with respect to an incident matrix.
    Neighborhood is defined on hypergraph level, not on adjacency level:
    neighbors of a node are edges, neighbors of an edge are nodes.

    Parameters
    ----------
    i: :class:`int`
        Index of the node/edge to probe.
    compressed_incidence: :class:`~scipy.sparse.csr_matrix` or :class:`~scipy.sparse.csc_matrix`
        Compressed sparse incidence matrix on rows (for nodes) or columns (for edges).

    Returns
    -------
    :class:`~numpy.ndarray`
        Neighbors of *i*.

    Examples
    --------

    A hypergraph with 4 nodes, 2 regular edges (0, 1) and (0, 2) and one 4-edge (0, 1, 2, 3).

    >>> import numpy as np
    >>> from scipy.sparse import csr_matrix, csc_matrix
    >>> incidence = np.array([[1, 1, 1],
    ...                       [1, 0, 1],
    ...                       [0, 1, 1],
    ...                       [0, 0, 1]])

    Edges of node 0:

    >>> neighbors(0, csr_matrix(incidence)) # doctest: +ELLIPSIS
    array([0, 1, 2]...)

    Egde of node 3:

    >>> neighbors(3, csr_matrix(incidence)) # doctest: +ELLIPSIS
    array([2]...)

    Nodes of edge 0:

    >>> neighbors(0, csc_matrix(incidence)) # doctest: +ELLIPSIS
    array([0, 1]...)

    Nodes of hyperedge 2:

    >>> neighbors(2, csc_matrix(incidence)) # doctest: +ELLIPSIS
    array([0, 1, 2, 3]...)
    """
    return compressed_incidence.indices[
        compressed_incidence.indptr[i] : compressed_incidence.indptr[i + 1]
    ]



from numba import njit

from stochastic_matching.simulator.simulator import Simulator
from stochastic_matching.simulator.multiqueue import MultiQueue


@njit
def fcfm_core(logs, arrivals, graph, n_steps, queue_size, queues):
    """
    Jitted function for first-come, first-matched policy.

    Parameters
    ----------
    logs: :class:`~stochastic_matching.simulator.logs.Logs`
        Monitored variables.
    arrivals: :class:`~stochastic_matching.simulator.arrivals.Arrivals`
        Item arrival process.
    graph: :class:`~stochastic_matching.simulator.graph.JitHyperGraph`
        Model graph.
    n_steps: :class:`int`
        Number of arrivals to process.
    queue_size: :class:`~numpy.ndarray`
        Number of waiting items of each class.
    queues: :class:`~stochastic_matching.simulator.multiqueue.MultiQueue`
        Waiting items of each class.

    Returns
    -------
    None
    """

    n, max_queue = logs.queue_log.shape
    inf = queues.infinity

    for age in range(n_steps):
        node = arrivals.draw()
        queue_size[node] += 1
        if queue_size[node] == max_queue:
            return None
        queues.add(node, age)

        # Test if an actionable edge may be present
        best_edge = -1
        if queue_size[node] == 1:
            best_age = inf
            # check oldest edge node
            for e in graph.edges(node):
                edge_age = inf
                for v in graph.nodes(e):
                    v_age = queues.oldest(v)
                    if v_age == inf:
                        break
                    edge_age = min(edge_age, v_age)
                else:
                    if edge_age < best_age:
                        best_edge = e
                        best_age = edge_age

            if best_edge > -1:
                # logs.traffic[best_edge] += 1
                queue_size[graph.nodes(best_edge)] -= 1
                for v in graph.nodes(best_edge):
                    queues.pop(v)

        logs.update(queue_size=queue_size, node=node, edge=best_edge)


class FCFM(Simulator):
    """
    Greedy Matching simulator derived from :class:`~stochastic_matching.simulator.simulator.Simulator`.
    When multiple choices are possible, the oldest item is chosen.

    Examples
    --------

    Let start with a working triangle. One can notice the results are the same for all greedy simulator because
    there are no multiple choices in a triangle (always one non-empty queue at most under a greedy policy).

    >>> import stochastic_matching as sm
    >>> sim = FCFM(sm.Cycle(rates=[3, 4, 5]), n_steps=1000, seed=42, max_queue=10)
    >>> sim.run()
    >>> sim.plogs # doctest: +NORMALIZE_WHITESPACE
    Arrivals: [287 338 375]
    Traffic: [125 162 213]
    Queues: [[838 104  41  13   3   1   0   0   0   0]
     [796 119  53  22   8   2   0   0   0   0]
     [640 176  92  51  24   9   5   3   0   0]]
    Steps done: 1000

    Unstable diamond (simulation ends before completion due to drift).

    >>> sim = FCFM(sm.CycleChain(rates=[1, 1, 1, 1]), n_steps=1000, seed=42, max_queue=10)
    >>> sim.run()
    >>> sim.plogs # doctest: +NORMALIZE_WHITESPACE
    Arrivals: [85 82 85 86]
    Traffic: [34 42  7 41 36]
    Queues: [[126  70  22  26  29  12  23  15  10   5]
     [326   8   3   1   0   0   0   0   0   0]
     [321  12   4   1   0   0   0   0   0   0]
     [105  80  65  28  31  15   4   2   6   2]]
    Steps done: 338

    A stable candy (but candies are not good for greedy policies).

    >>> sim = FCFM(sm.HyperPaddle(rates=[1, 1, 1.5, 1, 1.5, 1, 1]),
    ...            n_steps=1000, seed=42, max_queue=25)
    >>> sim.run()
    >>> sim.plogs # doctest: +NORMALIZE_WHITESPACE
    Arrivals: [46 26 32 37 70 35 45]
    Traffic: [24 17  2 23 33 12 13]
    Queues: [[ 23  32  45  38  22  43  31  34  20   3   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]
     [290   1   0   0   0   0   0   0   0   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]
     [290   1   0   0   0   0   0   0   0   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]
     [ 9   1   7   9   3   3  26  37   4   8  10   9   2  10  40  11   2  16
        3   3  21  27  22   1   7]
     [212  49  22   5   3   0   0   0   0   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]
     [233  41   6   7   4   0   0   0   0   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]
     [231  33  16   4   6   1   0   0   0   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]]
    Steps done: 291
    """

    name = "fcfm"

    def set_internal(self):
        super().set_internal()
        self.internal["queues"] = MultiQueue(self.model.n, max_queue=self.max_queue + 1)

    def run(self):
        fcfm_core(logs=self.logs, **self.internal)



from numba import njit
import numpy as np
from stochastic_matching.simulator.simulator import Simulator


@njit
def random_edge_selector(graph, queue_size, node):
    """
    Selects a feasible edge at random.
    """
    best_edge = -1
    best_score = -1.0
    for e in graph.edges(node):
        for v in graph.nodes(e):
            if queue_size[v] == 0:
                break
        else:
            score = np.random.rand()
            if score > best_score:
                best_score = score
                best_edge = e
    return best_edge


class RandomEdge(Simulator):
    """
    Greedy Matching simulator derived from :class:`~stochastic_matching.simulator.simulator.Simulator`.
    When multiple choices are possible, one edge is chosen uniformly at random.

    Parameters
    ----------
    model: :class:`~stochastic_matching.model.Model`
        Model to simulate.
    **kwargs
        Keyword arguments.


    Examples
    --------

    Let start with a working triangle. One can notice the results are not the than for other greedy simulators.
    This may seem off because there are no multiple choices in a triangle
    (always one non-empty queue at most under a greedy policy).
    The root cause is that the arrival process shares the same random generator than the random selection.

    >>> import stochastic_matching as sm
    >>> triangle = sm.Cycle(rates=[3, 4, 5])
    >>> sim = RandomEdge(triangle, n_steps=1000, seed=42, max_queue=11)
    >>> sim.run()
    >>> sim.plogs # doctest: +NORMALIZE_WHITESPACE
    Arrivals: [276 342 382]
    Traffic: [118 158 224]
    Queues: [[865  92  32  10   1   0   0   0   0   0   0]
     [750 142  62  28  12   3   2   1   0   0   0]
     [662 164  73  36  21   7  10  12   8   5   2]]
    Steps done: 1000

     Sanity check: results are unchanged if the graph is treated as hypergraph.

    >>> triangle.adjacency = None
    >>> sim = RandomEdge(triangle, n_steps=1000, seed=42, max_queue=11)
    >>> sim.run()
    >>> sim.plogs # doctest: +NORMALIZE_WHITESPACE
    Arrivals: [276 342 382]
    Traffic: [118 158 224]
    Queues: [[865  92  32  10   1   0   0   0   0   0   0]
     [750 142  62  28  12   3   2   1   0   0   0]
     [662 164  73  36  21   7  10  12   8   5   2]]
    Steps done: 1000

    An ill diamond graph (simulation ends before completion due to drift).

    >>> sim = RandomEdge(sm.CycleChain(rates='uniform'), n_steps=1000, seed=42, max_queue=10)
    >>> sim.run()
    >>> sim.plogs # doctest: +NORMALIZE_WHITESPACE
    Arrivals: [24 20 16 23]
    Traffic: [10  8  2  8  6]
    Queues: [[ 9  4  7 13  5  5 10 24  5  1]
     [83  0  0  0  0  0  0  0  0  0]
     [76  4  3  0  0  0  0  0  0  0]
     [13  8 15 19 14  8  2  1  2  1]]
    Steps done: 83
    >>> sim.flow
    array([0.48192771, 0.38554217, 0.09638554, 0.38554217, 0.28915663])

    A working candy (but candies are not good for greedy policies).

    >>> sim = RandomEdge(sm.HyperPaddle(rates=[1, 1, 1.5, 1, 1.5, 1, 1]), n_steps=1000, seed=42, max_queue=25)
    >>> sim.run()
    >>> sim.plogs # doctest: +NORMALIZE_WHITESPACE
    Arrivals: [47 43 67 50 92 47 54]
    Traffic: [24 22 19 28 35 19 26]
    Queues: [[305  77  15   3   0   0   0   0   0   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]
     [305  53  32   9   1   0   0   0   0   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]
     [341  36  18   5   0   0   0   0   0   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]
     [  7   6   9  13   7  34  41  47  23   9   2   3   2   7   6   2  12  46
       16  14  39  43   2   5   5]
     [275  51  38  30   6   0   0   0   0   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]
     [324  50  21   4   1   0   0   0   0   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]
     [305  47  23   9   2   3   5   5   1   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]]
    Steps done: 400

    Note that you can reset the simulator before starting another run.

    >>> sim.reset()
    >>> sim.run()
    >>> sim.plogs # doctest: +NORMALIZE_WHITESPACE
    Arrivals: [47 43 67 50 92 47 54]
    Traffic: [24 22 19 28 35 19 26]
    Queues: [[305  77  15   3   0   0   0   0   0   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]
     [305  53  32   9   1   0   0   0   0   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]
     [341  36  18   5   0   0   0   0   0   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]
     [  7   6   9  13   7  34  41  47  23   9   2   3   2   7   6   2  12  46
       16  14  39  43   2   5   5]
     [275  51  38  30   6   0   0   0   0   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]
     [324  50  21   4   1   0   0   0   0   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]
     [305  47  23   9   2   3   5   5   1   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]]
    Steps done: 400

    You can display the distribution of queue sizes as a ccdf:

    >>> sim.show_ccdf() # doctest: +ELLIPSIS
    <Figure size ...x... with 1 Axes>
    """

    name = "random_edge"

    def set_internal(self):
        super().set_internal()
        self.internal["selector"] = random_edge_selector



"""Top-level package for Stochastic Matching."""

from importlib.metadata import metadata
from stochastic_matching.model import Model
from stochastic_matching.graphs import (
    Path,
    Star,
    Cycle,
    Codomino,
    CycleChain,
    Complete,
    ErdosRenyi,
    NS19,
    Pyramid,
    HyperPaddle,
    KayakPaddle,
    Lollipop,
    Tadpole,
    Barbell,
    Fan,
    concatenate,
)
from stochastic_matching.simulator.simulator import Simulator
from stochastic_matching.simulator.fcfm import FCFM
from stochastic_matching.simulator.longest import Longest
from stochastic_matching.simulator.priority import Priority
from stochastic_matching.simulator.random_edge import RandomEdge
from stochastic_matching.simulator.random_item import RandomItem
from stochastic_matching.simulator.virtual_queue import VirtualQueue
from stochastic_matching.simulator.e_filtering import EFiltering
from stochastic_matching.simulator.constant_regret import ConstantRegret
from stochastic_matching.simulator.parallel import XP, Runner, Iterator, evaluate


infos = metadata(__name__)
__version__ = infos["Version"]
__author__ = """Fabien Mathieu"""
__email__ = "fabien.mathieu@normalesup.org"



import numpy as np
from numba import njit

from stochastic_matching.simulator.simulator import Simulator


def make_priority_selector(weights, threshold, counterweights):
    """
    Make a jitted edge selector based on priorities.
    """

    def priority_selector(graph, queue_size, node):
        best_edge = -1
        best_weight = -1000

        if threshold is None:
            w = weights
        else:
            for s in queue_size:
                if s >= threshold:
                    w = counterweights
                    break
            else:
                w = weights

        for e in graph.edges(node):
            for v in graph.nodes(e):
                if queue_size[v] == 0:
                    break
            else:
                if w[e] > best_weight:
                    best_weight = w[e]
                    best_edge = e
        return best_edge

    return njit(priority_selector)


class Priority(Simulator):
    """
    Greedy policy based on pre-determined preferences on edges.

    A threshold can be specified to alter the weights if the queue sizes get too big.

    Parameters
    ----------
    model: :class:`~stochastic_matching.model.Model`
        Model to simulate.
    weights: :class:`list` or :class:`~numpy.ndarray`
        Priorities associated to the edges.
    threshold: :class:`int`, optional
        Limit on max queue size to apply the weight priority.
    counterweights: :class:`list` or :class:`~numpy.ndarray`, optional
        Priority to use above threshold (if not provided, reverse weights is used).
    **kwargs
        Keyword arguments.

    Examples
    --------

    >>> import stochastic_matching as sm
    >>> fish = sm.KayakPaddle(m=4, l=0, rates=[4, 4, 3, 2, 3, 2])
    >>> fish.run('priority', weights=[0, 2, 2, 0, 1, 1, 0],
    ...                          threshold=50, counterweights = [0, 0, 0, 1, 2, 2, 1],
    ...                          n_steps=10000, seed=42)
    True

    These priorities are efficient at stabilizing the policy while avoiding edge 3.

    >>> fish.simulation
    array([2.925 , 1.0404, 0.9522, 0.    , 0.9504, 2.0808, 1.0044])

    The last node is the pseudo-instable node.

    >>> fish.simulator.avg_queues[-1]
    np.float64(38.346)
    >>> np.round(np.mean(fish.simulator.avg_queues[:-1]), decimals=2)
    np.float64(0.75)

    Choosing proper counter-weights is important.

    >>> fish.run('priority', weights=[0, 2, 2, 0, 1, 1, 0],
    ...                          threshold=50,
    ...                          n_steps=10000, seed=42)
    True
    >>> fish.simulation
    array([2.9232, 1.0422, 0.9504, 0.216 , 0.7344, 1.8666, 1.2186])
    >>> fish.simulator.avg_queues[-1]
    np.float64(38.6016)
    """

    name = "priority"

    def __init__(self, model, weights, threshold=None, counterweights=None, **kwargs):
        weights = np.array(weights)
        if threshold is not None:
            if counterweights is None:
                counterweights = -weights
            else:
                counterweights = np.array(counterweights)

        self.weights = weights
        self.threshold = threshold
        self.counterweights = counterweights

        super().__init__(model, **kwargs)

    def set_internal(self):
        super().set_internal()
        self.internal["selector"] = make_priority_selector(
            weights=self.weights,
            threshold=self.threshold,
            counterweights=self.counterweights,
        )



from numba import njit
import numpy as np
from stochastic_matching.simulator.simulator import Simulator


@njit
def random_item_selector(graph, queue_size, node):
    """
    Selects a feasible edge at random proportionally to the number of waiting items.
    """
    best_edge = -1
    prev_weight = 0.0

    for e in graph.edges(node):
        weight = 0.0
        for v in graph.nodes(e):
            w = queue_size[v]
            if w == 0:
                break
            weight += w
        else:
            prev_weight += weight
            if prev_weight * np.random.rand() < weight:
                best_edge = e
    return best_edge


class RandomItem(Simulator):
    """
    Greedy matching simulator derived from :class:`~stochastic_matching.simulator.simulator.Simulator`.
    When multiple choices are possible, chooses proportionally to the sizes of the queues
    (or sum of queues for hyperedges).

    Parameters
    ----------

    model: :class:`~stochastic_matching.model.Model`
        Model to simulate.
    **kwargs
        Keyword arguments.

    Examples
    --------

    Let start with a working triangle.

    >>> import stochastic_matching as sm
    >>> sim = RandomItem(sm.Cycle(rates=[3, 4, 5]), n_steps=1000, seed=42, max_queue=11)
    >>> sim.run()
    >>> sim.plogs # doctest: +NORMALIZE_WHITESPACE
    Arrivals: [276 342 382]
    Traffic: [118 158 224]
    Queues: [[865  92  32  10   1   0   0   0   0   0   0]
     [750 142  62  28  12   3   2   1   0   0   0]
     [662 164  73  36  21   7  10  12   8   5   2]]
    Steps done: 1000

    An ill braess graph (simulation ends before completion due to drift).

    >>> sim = RandomItem(sm.CycleChain(rates='uniform'), n_steps=1000, seed=42, max_queue=10)
    >>> sim.run()
    >>> sim.plogs # doctest: +NORMALIZE_WHITESPACE
    Arrivals: [25 17 13 12]
    Traffic: [10  6  2  5  5]
    Queues: [[ 9  4  7  7  5  4  9  8 11  3]
     [67  0  0  0  0  0  0  0  0  0]
     [60  4  3  0  0  0  0  0  0  0]
     [15 18 16  9  4  5  0  0  0  0]]
    Steps done: 67

    A working candy (but candies are not good for greedy policies).

    >>> sim = RandomItem(sm.HyperPaddle(rates=[1, 1, 1.5, 1, 1.5, 1, 1]), n_steps=1000, seed=42, max_queue=25)
    >>> sim.run()
    >>> sim.plogs # doctest: +NORMALIZE_WHITESPACE
    Arrivals: [112 106 166 105 206 112 115]
    Traffic: [66 46 39 61 64 51 81]
    Queues: [[683 125  63  25   8  13   2   3   0   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]
     [659 130  65  34  32   2   0   0   0   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]
     [757  89  36  19  12   9   0   0   0   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]
     [ 23  10  23  30  28  56  56 109  81  85 114  55  22  38  17  36  32  30
        7   4   3   5  13  20  25]
     [673 123  73  28  19   6   0   0   0   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]
     [769 112  33   4   4   0   0   0   0   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]
     [708 122  55  17   4   5   5   5   1   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]]
    Steps done: 922
    """

    name = "random_item"

    def set_internal(self):
        super().set_internal()
        self.internal["selector"] = random_item_selector



import json
import uuid
import numpy as np

from IPython.display import HTML, display

from stochastic_matching.common import neighbors


def int_2_str(model, i):
    """
    Parameters
    ----------
    model: :class:`~stochastic_matching.model.Model`
        A stochastic model.
    i: :class:`int`
        Node index.

    Returns
    -------
    :class:`str`
        Name of the node.

    Examples
    --------

    >>> import stochastic_matching as sm
    >>> diamond = sm.CycleChain()
    >>> int_2_str(diamond, 2)
    '2'
    >>> diamond.names = ['One', 'Two', 'Three', 'Four']
    >>> int_2_str(diamond, 2)
    'Three'
    >>> diamond.names = 'alpha'
    >>> int_2_str(diamond, 2)
    'C'
    """
    if model.names is None:
        return str(i)
    else:
        return model.names[i]


VIS_LOCATION = "https://unpkg.com/vis-network/standalone/umd/vis-network.min"
"""Default location of vis-network.js ."""

VIS_OPTIONS = {
    "interaction": {"navigationButtons": False},
    "width": "100%",
    "height": "100%",
}
"""Default options for the vis-network engine."""

HYPER_GRAPH_VIS_OPTIONS = {
    "groups": {
        "HyperEdge": {
            "fixed": {"x": False},
            "color": {"background": "black"},
            "shape": "dot",
            "size": 5,
        },
        "Node": {"fixed": {"x": False}},
    }
}
"""Default additional options for hypergraphs in the vis-network engine."""

# language=HTML
HTML_TEMPLATE = """
<div class="sm-graph" id="box-%(name)s">
<div id="%(name)s"></div>
<a href="#" id="fit-%(name)s"
  style="position: absolute; left: 10px; bottom: 10px; text-decoration: none; color: #888; font-size: min(2vw, 10px);
  z-index: 10; pointer-events: auto;"
> Reload
</a>
<style>
#box-%(name)s {
position: relative;
width: 100%%;
height: %(height)s !important;
max-width: 100vw;
max-height: 100vh !important;
}
#%(name)s {
  height: 100%%;   /* Make the inner div fill the parent */
}
</style>
</div>
<script type="module">
import { DataSet, Network } from "https://unpkg.com/vis-network/standalone/esm/vis-network.min.js";
// CSS will be automatically injected into the page.
function render () {
    const nodes = new DataSet(%(nodes)s);
    const edges = new DataSet(%(edges)s);
    const data = {
            nodes: nodes,
            edges: edges,
    };
    const options = %(options)s;
    const container = document.getElementById('%(name)s');
    const network = new Network(container, data, options);
    network.fit({maxZoomLevel: 10});
    network.redraw();
}

render();
document.getElementById('fit-%(name)s').addEventListener('click', function(event) {
    event.preventDefault();  // Prevent page jump
    render(); // Adjust zoom limit as needed
});
</script>
"""

OLD_HTML_TEMPLATE = """
<div class="sn-graph">
<div id="%(name)s"></div>
<a href="#" id="fit-%(name)s"
  style="position: absolute; left: 10px; bottom: 10px; text-decoration: none; color: #888; font-size: min(2vw, 10px);
  z-index: 10; pointer-events: auto;"
>
  &copy; Gismap 2025
</a>
</div>
<script type="module">
function loadVis(url, callback) {
    if (window.vis) {
        console.log("vis-network already loaded");
        callback();
        return;
    }
    // Lookup other loads in progress
    if (document.querySelector('script[data-vis-loaded]')) {
        // Script is loaded, wait until it is ready
        var check = setInterval(function () {
            if (window.vis) {
                clearInterval(check);
                console.log("vis-network sideload completed");
                callback();
            }
        }, 20);
        return;
    }
    // Add the script to the page, callback called when loaded
    var script = document.createElement('script');
    script.src = url;
    script.async = true;
    script.setAttribute('data-vis-loaded', '1');
    script.onload = callback;
    document.head.appendChild(script);
    console.log('vis-network dynamically loaded');
}
loadVis('%(vis)s', function () {
    const nodes = %(nodes)s;
    const edges = %(edges)s;
    const data = {
        nodes: nodes,
        edges: edges,
    };
    const options = %(options)s;
    const container = document.getElementById('%(name)s');
    const network = new vis.Network(container, data, options);
    network.fit({maxZoomLevel: 1});

    document.getElementById('fit-%(name)s').addEventListener('click', function(event) {
    event.preventDefault();  // Prevent page jump
    network.fit({ maxZoomLevel: 1 }); // Adjust zoom limit as needed
  });
})

"""

# language=HTML
FULL_HTML_TEMPLATE = """
<div id="%(name)s"></div>
<script type="module">
// 1. Define dark/light skins
const darkOptions = {
    nodes: {color: {background: "#222", border: "#999", highlight: "#444"}, font: {color: "#fff"}},
    edges: {color: {color: "#fff", highlight: "#FFD700"}, font: {color: "#fff"}}
};
const lightOptions = {
    nodes: {
        color: {
            background: "#fff",
            border: "#222",
            highlight: "#FFD700"},
        font: {color: "#222"}
    },
    edges: {color: {color: "#222", highlight: "#3876c2"}, font: {color: "#222"}}
};

// 2. Mode detection
function getTheme() {
    // Pydata Sphinx Theme: regarde la classe sur body
    const pydataTheme = document.documentElement.getAttribute("data-theme");
    if (pydataTheme === "dark" || pydataTheme === "light") {
        return pydataTheme;
    }
    // Jupyter
    const jupyterLabTheme = document.body.getAttribute("data-jp-theme-name");
    if (jupyterLabTheme) {
        // Simplify theme name to 'dark' or 'light'
        const lowerName = jupyterLabTheme.toLowerCase();
        if (lowerName.includes("dark")) {
            return "dark";
        }
        if (lowerName.includes("light")) {
            return "light";
        }
    }
    // System
    return window.matchMedia && window.matchMedia('(prefers-color-scheme: dark)').matches ? "dark" : "light";
}

// Universal function that loads vis.js only once
function loadVis(url, callback) {
    if (window.vis) {
        console.log("vis-network already loaded");
        callback();
        return;
    }
    // Lookup other loads in progress
    if (document.querySelector('script[data-vis-loaded]')) {
        // Script is loaded, wait until it is ready
        var check = setInterval(function () {
            if (window.vis) {
                clearInterval(check);
                console.log("vis-network sideload completed");
                callback();
            }
        }, 20);
        return;
    }
    // Add the script to the page, callback called when loaded
    var script = document.createElement('script');
    script.src = url;
    script.async = true;
    script.setAttribute('data-vis-loaded', '1');
    script.onload = callback;
    document.head.appendChild(script);
    console.log('vis-network dynamically loaded');
}

// Start display when vis is loaded
function render() {
loadVis('%(vis)s', function () {
    const nodes = %(nodes)s;
    const edges = %(edges)s;
    const data = {
        nodes: nodes,
        edges: edges,
    };
    const options = %(options)s;
    const container = document.getElementById('%(name)s');
    const network = new vis.Network(container, data, options);
    const theme = getTheme();
    container._theme = theme;
    const themeOptions = theme === "dark" ? darkOptions : lightOptions;
    network.setOptions(themeOptions);
    network.fit({maxZoomLevel: 200});
})}

render();
// Adapter dynamiquement si le thème change
window.addEventListener("theme-changed", () => render());
const observer = new MutationObserver(mutations => {
  for (const mutation of mutations) {
    if (mutation.type === "attributes" && mutation.attributeName === "data-jp-theme-name") {
      render();
    }
  }
    });
observer.observe(document.body, { attributes: true });
if (window.matchMedia) {
  window.matchMedia('(prefers-color-scheme: dark)').addEventListener('change', () => render());
}
</script>
"""
"""Default template."""

PNG_TEMPLATE = """
<div id="%(name)s"></div>
<img id="%(name)s-canvasImg" alt="Right click to save me!">
<script>
// Universal function that loads vis.js only once
function loadVis(url, callback) {
  if (window.vis) {
    console.log("vis-network already loaded");
    callback();
    return;
  }
  // Lookup other loads in progress
  if (document.querySelector('script[data-vis-loaded]')) {
    // Script is loaded, wait until it is ready
    var check = setInterval(function() {
      if (window.vis) {
        clearInterval(check);
        console.log("vis-network sideload completed");
        callback();
      }
    }, 20);
    return;
  }
  // Add the script to the page, callback called when loaded
  var script = document.createElement('script');
  script.src = url;
  script.async = true;
  script.setAttribute('data-vis-loaded', '1');
  script.onload = callback;
  document.head.appendChild(script);
  console.log('vis-network dynamically loaded');
}

// Start display when vis is loaded
loadVis('%(vis)s', function() {
  var nodes = %(nodes)s;
  var edges = %(edges)s;
  var data = {
    nodes: nodes,
    edges: edges,
  };
  var options = %(options)s;
  var container = document.getElementById('%(name)s');
  var network = new vis.Network(container, data, options);
  network.on("afterDrawing", function (ctx) {
      var dataURL = ctx.canvas.toDataURL();
      document.getElementById('%(name)s-canvasImg').src = dataURL;
    });
  network.fit({ maxZoomLevel: 1000 });
});
</script>
"""
"""Alternate template with a mirror PNG that ca be saved."""


def vis_code(
    vis_nodes=None,
    vis_edges=None,
    vis_options=None,
    template=None,
    vis=None,
    div_name=None,
):
    """
    Create HTML to display a Vis network graph.

    Parameters
    ----------
    vis_nodes: :class:`list` of :class:`dict`
        List the nodes of the graph. Each node is a dictionary with mandatory key `id`.
    vis_edges: :class:`list` of :class:`dict`
        List the edges of the graph. Each node is a dictionary with mandatory keys `from` and `to`.
    vis_options: :class:`dict`, optional
        Options to pass to Vis.
    template: :class:`str`, optional
        Template to use. Default to :obj:`~stochastic_matching.display.HTML_TEMPLATE`.
    vis: :class:`str`, optional
        Location of vis.js. Default to :obj:`~stochastic_matching.display.VIS_LOCATION`
    div_name: :class:`str`, optional
        Id of the div that will host the display.

    Returns
    -------
    :class:`str`
        Vis code (HTML by default).

    Examples
    --------
    >>> node_list = [{'id': 0}, {'id': 1}, {'id': 2}, {'id': 3}]
    >>> edge_list = [{'from': 0, 'to': 1}, {'from': 0, 'to': 2},
    ...          {'from': 1, 'to': 3}, {'from': 2, 'to': 3}]
    >>> print(vis_code(vis_nodes=node_list, vis_edges=edge_list)) # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
    <div class="sm-graph" id="box-...">
    <div id="..."></div>
    <a href="#" id="fit-..."
    style="position: absolute; left: 10px; bottom: 10px; text-decoration: none; color: #888; font-size: min(2vw, 10px);
    z-index: 10; pointer-events: auto;"
    > Reload
    </a>
    ...
    """
    if div_name is None:
        div_name = str(uuid.uuid4())
    if vis_nodes is None:
        vis_nodes = [{"id": 0}, {"id": 1}]
    if vis_edges is None:
        vis_edges = [{"from": 0, "to": 1}]
    if vis_options is None:
        vis_options = dict()
    if template is None:
        template = HTML_TEMPLATE
    if vis is None:
        vis = VIS_LOCATION
    vis_options = {**VIS_OPTIONS, **vis_options}
    height = vis_options["height"]
    if isinstance(height, int):
        height = f"{height}px"
        vis_options["height"] = height
    if "px" not in height:
        height = "50vh"
    dic = {
        "name": div_name,
        "nodes": json.dumps(vis_nodes),
        "edges": json.dumps(vis_edges),
        "options": json.dumps(vis_options),
        "height": height,
        "vis": vis,
    }
    return template % dic


def vis_show(
    vis_nodes=None,
    vis_edges=None,
    vis_options=None,
    template=None,
    vis=None,
    div_name=None,
):
    """
    Display a Vis graph (within a IPython / Jupyter session).

    Parameters
    ----------
    vis_nodes: :class:`list` of :class:`dict`
        List the nodes of the graph. Each node is a dictionary with mandatory key `id`.
    vis_edges: :class:`list` of :class:`dict`
        List the edges of the graph. Each node is a dictionary with mandatory keys `from` and `to`.
    vis_options: :class:`dict`, optional
        Options to pass to Vis.
    template: :class:`str`, optional
        Template to use. Default to :obj:`~stochastic_matching.display.HTML_TEMPLATE`.
    vis: :class:`str`, optional
        Location of vis.js. Default to :obj:`~stochastic_matching.display.VIS_LOCATION`
    div_name: :class:`str`, optional
        Id of the div that will host the display.

    Returns
    -------
    :class:`~IPython.display.HTML`

    Examples
    --------

    >>> vis_show()
    <IPython.core.display.HTML object>
    """
    # noinspection PyTypeChecker
    display(
        HTML(
            vis_code(
                vis_nodes=vis_nodes,
                vis_edges=vis_edges,
                vis_options=vis_options,
                template=template,
                vis=vis,
                div_name=div_name,
            )
        )
    )


def default_description(model):
    """
    Parameters
    ----------
    model: :class:`stochastic_matching.model.Model`
        Model to visualize.

    Returns
    -------
    nodes_info :class:`list` of `dict`
        Skeleton node description.
    nodes_info :class:`list` of `dict`
        Skeleton node description.

    Examples
    --------

    >>> import stochastic_matching as sm
    >>> diamond = sm.CycleChain()
    >>> default_description(diamond) # doctest: +NORMALIZE_WHITESPACE
    ([{'id': 0, 'label': '', 'title': '0'},
      {'id': 1, 'label': '', 'title': '1'},
      {'id': 2, 'label': '', 'title': '2'},
      {'id': 3, 'label': '', 'title': '3'}],
    [{'title': '0: (0, 1)', 'label': ''},
     {'title': '1: (0, 2)', 'label': ''},
     {'title': '2: (1, 2)', 'label': ''},
     {'title': '3: (1, 3)', 'label': ''},
     {'title': '4: (2, 3)', 'label': ''}])
    >>> diamond.names = 'alpha'
    >>> default_description(diamond) # doctest: +NORMALIZE_WHITESPACE
    ([{'id': 0, 'label': '', 'title': '0: A'},
      {'id': 1, 'label': '', 'title': '1: B'},
      {'id': 2, 'label': '', 'title': '2: C'},
      {'id': 3, 'label': '', 'title': '3: D'}],
    [{'title': '0: (A, B)', 'label': ''},
     {'title': '1: (A, C)', 'label': ''},
     {'title': '2: (B, C)', 'label': ''},
     {'title': '3: (B, D)', 'label': ''},
     {'title': '4: (C, D)', 'label': ''}])
    """
    nodes_info = [
        {
            "id": i,
            "label": "",
            "title": f"{i}: {int_2_str(model, i)}"
            if model.names is not None
            else str(i),
        }
        for i in range(model.n)
    ]
    edges_info = [
        {"title": f"{j}: ({', '.join([int_2_str(model, i) for i in e])})", "label": ""}
        for j, e in [(j, neighbors(j, model.incidence_csc)) for j in range(model.m)]
    ]
    return nodes_info, edges_info


def vis_converter(model, nodes_info, edges_info):
    """
    Parameters
    ----------
    model: :class:`~stochastic_matching.model.Model`
        Model to visualize.
    nodes_info: :class:`list` of :class:`dict`
        Description of nodes.
    edges_info: :class:`list` of :class:`dict`
        Description of edges.

    Returns
    -------
    vis_nodes: :class:`list` of :class:`dict`
        Description of the nodes that will be displayed in vis.
        If the graph is simple, this is just the input nodes.
        For hypergraphs, both nodes and hyperedges are displayed as nodes in vis.
    vis_edges: :class:`list` of :class:`dict`
        Description of the edges that will be displayed in vis.
        If the graph is simple, this is just the input edges, with endpoints info added for vis.
        For hypergraphs, each edge in vis links a node and a hyperedge..


    Examples
    --------
    >>> import stochastic_matching as sm
    >>> diamond = sm.CycleChain()
    >>> nodes, edges = default_description(diamond)
    >>> vis_converter(diamond, nodes, edges) # doctest: +NORMALIZE_WHITESPACE
    ([{'id': 0, 'label': '', 'title': '0'},
      {'id': 1, 'label': '', 'title': '1'},
      {'id': 2, 'label': '', 'title': '2'},
      {'id': 3, 'label': '', 'title': '3'}],
     [{'title': '0: (0, 1)', 'label': '', 'from': 0, 'to': 1},
      {'title': '1: (0, 2)', 'label': '', 'from': 0, 'to': 2},
      {'title': '2: (1, 2)', 'label': '', 'from': 1, 'to': 2},
      {'title': '3: (1, 3)', 'label': '', 'from': 1, 'to': 3},
      {'title': '4: (2, 3)', 'label': '', 'from': 2, 'to': 3}])
    >>> diamond.adjacency = None
    >>> vis_converter(diamond, nodes, edges) # doctest: +NORMALIZE_WHITESPACE
    ([{'id': 0, 'label': '', 'title': '0', 'x': 0, 'group': 'Node'},
      {'id': 1, 'label': '', 'title': '1', 'x': 0, 'group': 'Node'},
      {'id': 2, 'label': '', 'title': '2', 'x': 0, 'group': 'Node'},
      {'id': 3, 'label': '', 'title': '3', 'x': 0, 'group': 'Node'},
      {'title': '0: (0, 1)', 'label': '', 'from': 0, 'to': 1, 'id': 4, 'group': 'HyperEdge', 'x': 600},
      {'title': '1: (0, 2)', 'label': '', 'from': 0, 'to': 2, 'id': 5, 'group': 'HyperEdge', 'x': 600},
      {'title': '2: (1, 2)', 'label': '', 'from': 1, 'to': 2, 'id': 6, 'group': 'HyperEdge', 'x': 600},
      {'title': '3: (1, 3)', 'label': '', 'from': 1, 'to': 3, 'id': 7, 'group': 'HyperEdge', 'x': 600},
      {'title': '4: (2, 3)', 'label': '', 'from': 2, 'to': 3, 'id': 8, 'group': 'HyperEdge', 'x': 600}],
     [{'from': 0, 'to': 4, 'title': '0 <-> 0: (0, 1)'},
      {'from': 0, 'to': 5, 'title': '0 <-> 1: (0, 2)'},
      {'from': 1, 'to': 4, 'title': '1 <-> 0: (0, 1)'},
      {'from': 1, 'to': 6, 'title': '1 <-> 2: (1, 2)'},
      {'from': 1, 'to': 7, 'title': '1 <-> 3: (1, 3)'},
      {'from': 2, 'to': 5, 'title': '2 <-> 1: (0, 2)'},
      {'from': 2, 'to': 6, 'title': '2 <-> 2: (1, 2)'},
      {'from': 2, 'to': 8, 'title': '2 <-> 4: (2, 3)'},
      {'from': 3, 'to': 7, 'title': '3 <-> 3: (1, 3)'},
      {'from': 3, 'to': 8, 'title': '3 <-> 4: (2, 3)'}])
    >>> candy = sm.HyperPaddle()
    >>> vis_nodes, vis_edges = vis_converter(candy, *default_description(candy))
    >>> vis_nodes[2]
    {'id': 2, 'label': '', 'title': '2', 'x': 0, 'group': 'Node'}
    >>> vis_nodes[13]
    {'title': '6: (2, 3, 4)', 'label': '', 'id': 13, 'group': 'HyperEdge', 'x': 600}
    >>> vis_edges[6]
    {'from': 2, 'to': 13, 'title': '2 <-> 6: (2, 3, 4)'}
    """
    simple = model.adjacency is not None
    if simple:
        vis_nodes = nodes_info
        vis_edges = edges_info
        for e, dico in enumerate(vis_edges):
            endpoints = neighbors(e, model.incidence_csc)
            dico["from"] = int(endpoints[0])
            dico["to"] = int(endpoints[1])
        return vis_nodes, vis_edges
    else:
        vis_nodes_1 = nodes_info
        vis_nodes_2 = edges_info
        for d in vis_nodes_1:
            d["x"] = 0
            d["group"] = "Node"
        for j, d in enumerate(vis_nodes_2):
            d["id"] = model.n + j
            d["group"] = "HyperEdge"
            d["x"] = 600
        vis_edges = [
            {
                "from": i,
                "to": model.n + int(j),
                "title": f"{vis_nodes_1[i]['title']} <-> {vis_nodes_2[j]['title']}",
            }
            for i in range(model.n)
            for j in neighbors(i, model.incidence_csr)
        ]
        return vis_nodes_1 + vis_nodes_2, vis_edges


def info_maker(
    model,
    disp_rates=True,
    disp_flow=True,
    flow=None,
    disp_kernel=False,
    disp_zero=True,
    check_flow=False,
    check_tolerance=1e-2,
):
    """
    Parameters
    ----------
    model: :class:`~stochastic_matching.model.Model`
        Model to visualize.
    disp_rates: :class:`bool`, optional
        Labels the nodes with their rates. Otherwise the names are used.
    disp_flow: :class:`bool`, optional
        Label the edges with the given flow.
    flow: :class:`~numpy.ndarray`, optional
        Flow to use. If None, the base flow will be used. If not None, overrides `disp_flow`.
    disp_kernel: :class:`bool`, optional
        Display the kernel basis on the edges. Compatible with the display of a flow.
    disp_zero: :class:`bool`, optional
        If False, do not label the edge with a null flow.
    check_flow: :class:`bool`, optional
        If True, color the edges with their positivity and the nodes with their compliance to the conservation law.
    check_tolerance: :class:`float`, optional
        Relative error when checking conservation law on nodes.
        For simulations, a relatively high value is recommended, for example 1e-2.

    Returns
    -------
    nodes_info: :class:`list` of :class:`dict`
        Description of nodes.
    edges_info: :class:`list` of :class:`dict`
        Description of edges.

    Examples
    ---------

    It is probably best to play a bit with the options, but the following examples should give the general idea.

    We start with the so-called *pyramid* graph (the names comes from one of its kernel,
    and not from the shape of the graph itself).

    >>> import stochastic_matching as sm
    >>> pyramid = sm.Pyramid(names='alpha')

    By default, the label of a node (its displayed name) is its arrival rate,
    and the label of an edge is its Moore_penrose flow.

    >>> n_i, e_i = info_maker(pyramid)
    >>> n_i[3]
    {'id': 3, 'label': '2', 'title': '3: D'}
    >>> e_i[2]
    {'title': '2: (B, C)', 'label': '1'}

    We can disable the display of the arrival rates, so the actual name of the node will be displayed.

    >>> n_i, e_i = info_maker(pyramid, disp_rates=False)
    >>> n_i[3]
    {'id': 3, 'label': 'D', 'title': '3: D'}

    We ask for no label on edges.

    >>> n_i, e_i = info_maker(pyramid, disp_flow=False)
    >>> e_i[2]
    {'title': '2: (B, C)', 'label': ''}

    We can set custom weights on the edges, for instance use a different flow vector.

    >>> flow = np.array([0., 2., 0., 3., 1., 1., 0., 2., 0., 2., 0., 1., 1.])
    >>> n_i, e_i = info_maker(pyramid, flow=flow)
    >>> assert np.allclose([float(e_i[i]['label']) for i in range(pyramid.m)], flow)

    We can ask for the kernel basis to be indicated as well.

    >>> n_i, e_i = info_maker(pyramid, flow=flow, disp_kernel=True)
    >>> e_i[2]
    {'title': '2: (B, C)', 'label': '0+α1'}
    >>> e_i[3]
    {'title': '3: (B, F)', 'label': '3-α1+α2-α3'}

    We can remove the flow to have just the kernel basis.

    >>> n_i, e_i = info_maker(pyramid, disp_flow=False, disp_kernel=True)
    >>> e_i[2]
    {'title': '2: (B, C)', 'label': '+α1'}
    >>> e_i[3]
    {'title': '3: (B, F)', 'label': '-α1+α2-α3'}

    By asking null values to be silent, we get avoid things like `0+...`.

    >>> n_i, e_i = info_maker(pyramid, flow=flow, disp_zero=False, disp_kernel=True)
    >>> e_i[2]
    {'title': '2: (B, C)', 'label': '+α1'}
    >>> e_i[3]
    {'title': '3: (B, F)', 'label': '3-α1+α2-α3'}

    If we ask to check the flow, null edges are displayed in orange.

    >>> n_i, e_i = info_maker(pyramid, flow=flow, check_flow=True)
    >>> e_i[2]
    {'title': '2: (B, C)', 'label': '0', 'color': 'orange'}

    Note that the kernel basis is not necessarily +/- 1, even on simple graphs.

    >>> kayak = sm.KayakPaddle()
    >>> n_i, e_i = info_maker(kayak, disp_kernel=True)
    >>> e_i[3]
    {'title': '3: (2, 3)', 'label': '1+2α1'}

    When kernel is displayed, edges that are not part of any kernel are shown in black.

    >>> diamond = sm.CycleChain()
    >>> n_i, e_i = info_maker(diamond, disp_kernel=True)
    >>> e_i[2]
    {'title': '2: (1, 2)', 'label': '1', 'color': 'black'}

    Nodes that do not check the conservation law and negative edges are shown in red.

    >>> n_i, e_i = info_maker(diamond, flow=[-1]*5, check_flow=True)
    >>> n_i[2]
    {'id': 2, 'label': '3', 'title': '2', 'color': 'red'}
    >>> e_i[2]
    {'title': '2: (1, 2)', 'label': '-1', 'color': 'red'}
    """
    nodes_info, edges_info = default_description(model)
    if flow is not None:
        disp_flow = True
    for i, node in enumerate(nodes_info):
        if disp_rates:
            node["label"] = f"{model.rates[i]:.3g}"
        else:
            node["label"] = int_2_str(model, i)
    if disp_flow:
        if flow is None:
            flow = model.base_flow
        for e, edge in enumerate(edges_info):
            if np.abs(flow[e]) > model.tol:
                edge["label"] = f"{flow[e]:.3g}"
            elif disp_zero:
                edge["label"] = "0"
    if disp_kernel:
        d, m = model.kernel.right.shape
        for e, edge in enumerate(edges_info):
            label = ""
            for i in range(d):
                alpha = model.kernel.alt_right[i, e]
                if alpha == 0:
                    continue
                if alpha == 1:
                    label += f"+"
                elif alpha == -1:
                    label += f"-"
                else:
                    label += f"{alpha:+.3g}"
                label += f"α{i + 1}"
            edge["label"] += label
            if not label:
                edge["color"] = "black"
    if check_flow and disp_flow:
        out_rate = model.incidence_csr @ flow
        for i, node in enumerate(nodes_info):
            if np.abs(model.rates[i] - out_rate[i]) / model.rates[i] > check_tolerance:
                node["color"] = "red"
        for e, edge in enumerate(edges_info):
            if flow[e] < -model.tol:
                edge["color"] = "red"
            elif flow[e] < model.tol:
                edge["color"] = "orange"
            else:
                edge["color"] = edge.get("color", "blue")
    return nodes_info, edges_info


def show(model, bipartite=False, png=False, **kwargs):
    """
    End-to-end display solution for model.
    It is basically a pipe
    :class:`~stochastic_matching.model.Model` ->
    :meth:`~stochastic_matching.display.info_maker` ->
    :meth:`~stochastic_matching.display.vis_converter` ->
    :meth:`~stochastic_matching.display.vis_show`.

    The extra arguments are passed when needed on the right spot along the pipe, allowing maximal flexibility.

    Parameters
    ----------
    model: :class:`~stochastic_matching.model.Model`
        The model to display.
    bipartite: :class:`bool`, optional
        Tells if the bipartite node/edge structure should be explicitly shown.
    png: :class:`bool`
        Make a mirror PNG that can be saved.
    **kwargs
        Keyword arguments. See :meth:`~stochastic_matching.display.info_maker`,
        :meth:`~stochastic_matching.display.vis_converter`, and
        :meth:`~stochastic_matching.display.vis_show` for details.


    Returns
    -------
    :class:`~IPython.display.HTML`

    Examples
    --------

    >>> import stochastic_matching as sm
    >>> pyramid = sm.Pyramid()

    Basic display.

    >>> show(pyramid)
    <IPython.core.display.HTML object>

    With this, the nodes and edges will not show anything. The result can be exported as png.

    >>> nodes_info=[{'label': ''} for _ in range(pyramid.n)]
    >>> edges_info=[{'label': ''} for _ in range(pyramid.m)]
    >>> show(pyramid, nodes_info=nodes_info, edges_info=edges_info, png=True)
    <IPython.core.display.HTML object>

    Fan is a true hypergraph.

    >>> fan = sm.Fan()

    To display:

    >>> show(fan)
    <IPython.core.display.HTML object>

    To display in bipartite mode:

    >>> show(fan, bipartite=True)
    <IPython.core.display.HTML object>
    """
    info_kwds = {
        "disp_rates",
        "disp_flow",
        "flow",
        "disp_kernel",
        "disp_zero",
        "check_flow",
        "check_tolerance",
    }
    converter_kwds = {"nodes_info", "edges_info"}
    vis_kwds = {"vis_options", "template", "vis", "div_name"}

    info_kwargs = {k: v for k, v in kwargs.items() if k in info_kwds}
    converter_kwargs = {k: v for k, v in kwargs.items() if k in converter_kwds}
    vis_kwargs = {k: v for k, v in kwargs.items() if k in vis_kwds}

    nodes_info, edges_info = info_maker(model, **info_kwargs)

    if "nodes_info" in converter_kwargs:
        nodes_info = [
            {**n1, **n2} for n1, n2 in zip(nodes_info, converter_kwargs["nodes_info"])
        ]
    if "edges_info" in converter_kwargs:
        edges_info = [
            {**e1, **e2} for e1, e2 in zip(edges_info, converter_kwargs["edges_info"])
        ]

    vis_nodes, vis_edges = vis_converter(model, nodes_info, edges_info)

    if model.adjacency is None:
        vis_kwargs["vis_options"] = {
            **vis_kwargs.get("vis_options", dict()),
            **HYPER_GRAPH_VIS_OPTIONS.copy(),
        }
        vis_options = vis_kwargs["vis_options"]
        if bipartite:
            vis_options["groups"]["HyperEdge"]["fixed"]["x"] = True
            vis_options["groups"]["Node"]["fixed"]["x"] = True
            inner_width = round(0.8 * vis_options.get("width", 600))
            for vis_node in vis_nodes:
                if vis_node.get("group") == "HyperEdge":
                    vis_node["x"] = inner_width
        else:
            vis_options["groups"]["HyperEdge"]["fixed"]["x"] = False
            vis_options["groups"]["Node"]["fixed"]["x"] = False
    if png:
        vis_kwargs["template"] = PNG_TEMPLATE
    vis_show(vis_nodes, vis_edges, **vis_kwargs)
